from typing import Dict


class Atom:
    """Atom is a reusable action.

    This class is a marker node and also performs basic dsl functions.
    """

    def then(self, other):
        from baseten.workflow.worklet import wrap
        return wrap(self).then(other)

    def __rshift__(self, other):
        return self.then(other)

    @property
    def node_name(self):
        raise NotImplementedError('Derived classes should override.')


class RegisteredAtom(Atom):
    """RegisteredAtom is a registered reusable action."""

    def __init__(self, name: str, node_name: str = '', conf=None):
        self._conf = conf or {}
        self._name = name
        self._node_name = node_name

    @property
    def conf(self) -> Dict:
        return self._conf

    @property
    def name(self) -> str:
        return self._name

    @property
    def node_name(self):
        return self._node_name



import copy
import json
import logging
import re
import tempfile
import zipfile
from functools import reduce
from typing import IO, Any, List

import joblib
import requests
from colorama import Fore, Style

import baseten
from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api
from baseten.common.core import (FrameworkNotSupportedError,
                                 ModelInputShapeError, ModelNotSupportedError,
                                 raises_api_error)
from baseten.common.settings import API_URL_BASE, SKLEARN, TENSORFLOW
from baseten.common.util import (infer_sklearn_packages,
                                 infer_tensorflow_packages, zipdir)

logger = logging.getLogger(__name__)


@raises_api_error
def deploy_model(serialized_model: IO[Any],
                 model_framework: str,
                 model_name: str = None,
                 model_type: str = None,
                 input_shape: List[int] = None,
                 feature_names: List[str] = None,
                 feature_summary: dict = None,
                 class_labels: List[str] = None,
                 model_framework_req: List[str] = None) -> BasetenDeployedModel:
    """Create a model version based on the serialized_model. If baseten wasn't initialized with a model_id, then
    create a model first and then create the model version.

    Args:
        serialized_model (file): A file-like object that is the serialized representation of the model object.
        model_framework (str): The framework used to create the model.
        model_name (str, optional): The name of the model to be created, if necessary.
        model_type (str, optional): The type of the model to be created.
        input_shape (List[int], optional): The shape of the model's input; e.g: [10] for a model with 10 features;
            [28, 28] for a model with 2 input dimensions of size 28 each.
        feature_names (list, optional): The list of feature names.
        feature_summary (dict, optional): A dict of summary stats with feature names as keys.
        class_labels (List[str], optional): The list of target class labels, if applicable.
        model_framework_req (List[str], optional): The model_framework specific requirements
    Returns:
        BasetenDeployedModel

    Raises:
        ApiError: If there was an error communicating with the server.
    """
    if not feature_summary:
        feature_summary = {}
    if not class_labels:
        class_labels = []
    if not model_framework_req:
        model_framework_req = {}
    if not baseten.working_model_id:
        # Create a new model

        logger.info(f'Making contact with BaseTen 👋 👽')
        model_json = api.create_model(model_name)
        logger.debug(f'Created model:\n{json.dumps(model_json, indent=4)}')
        logger.info(f'Successfully registered model {Fore.BLUE}{model_name}{Style.RESET_ALL}.')
        baseten.working_model_id = model_json['id']

    # Upload the model
    file_ext = _model_ext_from_framework(model_framework)
    model_key = _upload_model(serialized_model, file_ext)
    # Create the model version
    logger.info(f'{Fore.BLUE}Deploying model version.')
    model_version_json = api.create_model_version(baseten.working_model_id,
                                                  model_key,
                                                  model_framework,
                                                  model_type,
                                                  input_shape,
                                                  feature_names,
                                                  feature_summary,
                                                  class_labels,
                                                  model_framework_req)
    model_version_id = model_version_json['id']
    model_version_web_url = f'{API_URL_BASE}/models/{baseten.working_model_id}/versions/{model_version_id}'
    logger.debug(f'Created model version:\n{json.dumps(model_version_json, indent=4)}')
    logger.info(f'🏁 The model is being deployed right now 🏁')

    visit_message = f'|  Visit {Fore.BLUE}{model_version_web_url}{Style.RESET_ALL} for deployment status  |'
    visit_message_len = len(visit_message) - len(Fore.BLUE) - len(Style.RESET_ALL)
    logger.info(''.join(['-' for _ in range(visit_message_len)]))
    logger.info(visit_message)
    logger.info(''.join(['-' for _ in range(visit_message_len)]))

    return BasetenDeployedModel(model_version_id=model_version_id, model_name=model_name)


@raises_api_error
def deploy_model_object(
    model: Any,
    model_name: str = None,
    feature_names: List[str] = None,
    class_labels: List[str] = None
) -> BasetenDeployedModel:
    """Create a model version based on the model. If baseten wasn't initialized with a model_id, then
    create a model first and then create the model version.

    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a RandomForestClassifier object)
        model_name (str, optional): The name of the model to be created, if necessary.
        feature_names (List[str], optional): The list of feature names. If not provided,
            a list of feature names that match the model's input dimensions will be automatically generated.
            If the model's input has multiple dimensions, feature_names should be a flattened list of names.
        class_labels (List[str], optional): The list of target class labels, if applicable.
    Returns:
        BasetenDeployedModel

    Raises:
        FrameworkNotSupportedError: If the model framework used is not supported.
        ModelInputShapeError: feature_names does not match the model's signature.
        ApiError: If there was an error communicating with the server.
    """
    if hasattr(model, 'baseten'):
        model = copy.deepcopy(model)
        del model.baseten
    logger.info(f'🤖 Extracting 🤖 model 🤖 metadata 🤖')
    model_class = model.__class__
    model_framework, _, _ = model_class.__module__.partition('.')
    model_type = model_class.__name__
    input_shape = _model_input_shape(model, model_framework)
    logger.info(f'Preparing {Fore.BLUE}{model_framework} - {model_type}{Style.RESET_ALL} model.')

    if feature_names:
        input_features_count = reduce(lambda x, y: x * y, input_shape)
        if len(feature_names) != input_features_count:
            raise ModelInputShapeError(f'The model has {input_features_count} features but feature_names provided '
                                       f'{len(feature_names)}')

    if model_framework == SKLEARN:
        logger.info(f'Serializing {Fore.BLUE}{model_name}{Style.RESET_ALL}.')
        serialized_model = _serialize_sklearn_model(model)
        model_framework_req = infer_sklearn_packages()
        deployed_model = deploy_model(serialized_model,
                                      model_framework,
                                      model_name,
                                      model_type,
                                      input_shape,
                                      feature_names=feature_names,
                                      class_labels=class_labels,
                                      model_framework_req=model_framework_req)
        serialized_model.close()
    elif model_framework == TENSORFLOW:
        logger.info(f'Serializing {Fore.BLUE}{model_name}{Style.RESET_ALL}.')
        serialized_model = _serialize_tf_model(model)
        model_framework_req = infer_tensorflow_packages()
        deployed_model = deploy_model(serialized_model,
                                      model_framework,
                                      model_name,
                                      model_type,
                                      input_shape,
                                      feature_names=feature_names,
                                      class_labels=class_labels,
                                      model_framework_req=model_framework_req)
        serialized_model.close()
    else:
        raise FrameworkNotSupportedError(f'Models derived from {model_framework} are not currently supported.')

    if deployed_model:
        deployed_model.generate_model_template(
            model_type,
            model_framework,
            feature_names
        )
    return deployed_model


def _upload_model(serialized_model: IO, file_ext: str) -> str:
    """Uploads the serialized model to the appropriate environment

    Args:
        serialized_model (file): A file-like object that is the serialized representation of the model object.
        file_ext (str): The file extension for the saved model

    Returns:
        str: The key for the uploaded model

    Raises:
        RequestException: If there was an error communicating with the server.
    """

    model_file_name = f'model.{file_ext}'
    signed_s3_upload_post = api.signed_s3_upload_post(baseten.working_model_id, model_file_name)
    logger.debug(f'Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}')

    # Upload the serialized model to S3
    files = {'file': (model_file_name, serialized_model)}
    form_fields = signed_s3_upload_post['form_fields']
    form_fields['AWSAccessKeyId'] = form_fields.pop('aws_access_key_id')  # S3 expects key name AWSAccessKeyId
    logger.info(f'🚀 Uploading model to BaseTen 🚀')
    resp = requests.post(signed_s3_upload_post['url'], data=form_fields, files=files)
    logger.info(f'🖖 Upload success! 🖖')

    logger.debug(f'File upload HTTP status code: {resp.status_code} and content:\n{resp.content}')
    resp.raise_for_status()
    return signed_s3_upload_post['form_fields']['key']


def _serialize_sklearn_model(model: Any) -> tempfile.SpooledTemporaryFile:
    """Serializes a model based on the scikit-learn framework.

    Args:
        model (Any): The model object.

    Returns:
        tempfile.SpooledTemporaryFile: A temporary file wrapper.
    """
    model_joblib = tempfile.TemporaryFile()
    joblib.dump(model, model_joblib, compress=True)
    model_joblib.seek(0)
    return model_joblib


def _serialize_tf_model(model: Any) -> tempfile.NamedTemporaryFile:
    """Serializes a Tensorflow model as a zipped SavedModel.

    Args:
        model (Any): The model object.

    Returns:
        tempfile.NamedTemporaryFile: A temporary zipfile.ZipFile wrapper.
    """
    temp_model_dir = tempfile.TemporaryDirectory()
    temp_file = tempfile.NamedTemporaryFile(suffix='.zip')
    model.save(temp_model_dir.name)
    zip_file = zipfile.ZipFile(temp_file.name, 'w', zipfile.ZIP_DEFLATED)
    zipdir(temp_model_dir.name, zip_file)
    zip_file.close()
    temp_model_dir.cleanup()
    temp_file.file.seek(0)
    return temp_file


def _model_ext_from_framework(model_framework: str) -> str:
    if model_framework == SKLEARN:
        return 'joblib'
    elif model_framework == TENSORFLOW:
        return 'zip'
    else:
        raise FrameworkNotSupportedError(f'Models derived from {model_framework} are not currently supported.')


def _model_input_shape(model, framework: str) -> List[int]:
    """Returns the shape of the model's input as an array; e.g: [10] for a model with 10 features; [28, 28] for a
    model with 2 input dimensions of size 28 each.
    """
    if framework == SKLEARN:
        return [_hack_n_features_from_sklearn_model(model)]
    elif framework == TENSORFLOW:
        input_shape = model.input_shape
        if isinstance(input_shape, list):
            raise ModelNotSupportedError("Baseten currently supports only single input models.")
        return list(input_shape[1:])  # The first dimension is None (batch size)
    else:
        raise FrameworkNotSupportedError(f'Models derived from {framework} are not currently supported.')


def _hack_n_features_from_sklearn_model(model) -> int:
    """Determines the size of the input for the given model.

    Due to inconsistent representation of feature inputs in sklearn, various methods/props
    are required to extract the input dimension based on the model type. It's a hack.

    Args:
        model: A sklearn model object.

    Returns:
        int: The number of features used to train the model.
    """
    if 'sklearn.neighbors' in str(model.__class__):
        return model._tree.data.shape[-1]
    if 'sklearn.naive_bayes.GaussianNB' in str(model.__class__):
        return model.theta_.shape[-1]
    if model.__module__ == 'sklearn.naive_bayes':
        return model.feature_log_prob_.shape[-1]
    if model.__module__ == 'sklearn.discriminant_analysis':
        return model.means_.shape[-1]
    if 'sklearn.gaussian_process' in model.__module__:
        return model.base_estimator_.X_train_.shape[-1]
    try:
        model.predict([[1]])
        n_features = 1
    except ValueError as e:
        # For many models (not the ones above), the correct input feature length is raised in the exception
        numbers = re.findall(r' (\d+)', str(e))
        n_features = [int(x) for x in numbers if x != '1'][-1]
    return n_features



"""Baseten

    isort:skip_file
"""

__version__ = '0.1.0'

import logging
import os
from collections import defaultdict

patched_modules = defaultdict(lambda: [])

from baseten.baseten_deployed_model import BasetenDeployedModel  # noqa: E402
from baseten.baseten_deployed_model import install_requirements, requirements_status  # noqa: E402, F401
from baseten.common.external_connections import add_external_connection, external_connections  # noqa: E402, F401
from baseten.common.model_deployer import deploy_model_object  # noqa: E402
from baseten.common.core import BASETEN_CONFIG_DIRECTORY  # noqa: E402
from baseten.common.util import setup_logger  # noqa: E402
from baseten.common import settings  # noqa: E402
from baseten.workflows import BasetenWorkflow  # noqa: E402


logger = logging.getLogger(__name__)
setup_logger('baseten', logging.INFO)
if settings.DEBUG:
    setup_logger('baseten', logging.DEBUG)
logger.debug(f'Starting the client with the server URL set to {settings.API_URL_BASE}')

# The Baseten model ID that either a) the user initialized baseten with, or b) was created when deploying a model.
working_model_id = None
deploy = deploy_model_object  # This allows the user to call baseten.deploy(model)


def set_log_level(level):
    setup_logger('baseten', level)


def init(baseten_model_id=None):
    """Initialize Baseten

    Args:
        baseten_model_id (str, optional): The BaseTen model id to initialize the client with.
        If not provided, a new Baseten model will be created at deploy() time.
    """
    global working_model_id
    working_model_id = baseten_model_id

    if not os.path.exists(BASETEN_CONFIG_DIRECTORY):
        os.makedirs(BASETEN_CONFIG_DIRECTORY)


def deployed_model_id(model_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model model_id.

    Args:
        model_id (str)

    Returns:
        BasetenDeployedModel
    """
    return BasetenDeployedModel(model_id=model_id)


def deployed_model_version_id(model_version_id: str) -> BasetenDeployedModel:
    """Returns a BasetenDeployedModel object for interacting with the model version model_version_id.

    Args:
        model_version_id (str)

    Returns:
        BasetenDeployedModel
    """
    return BasetenDeployedModel(model_version_id=model_version_id)


def workflow_by_id(workflow_id: str) -> BasetenWorkflow:
    """Returns a BasetenWorkflow object for interacting with the workflow.

    Args:
        workflow_id (str)

    Returns:
        BasetenWorkflow
    """
    return BasetenWorkflow(workflow_id=workflow_id)



from jinja2 import Template

model_template = Template(
    """
model:
{% if model_type %}
    model_type: {{ model_type }}
{% endif %}
{% if model_framework %}
    model_framework: {{ model_framework }}
{% endif %}
model_features:
    input_type: {{ input_type }}
    features:
{% for feature_name in feature_names %}
        - {{ feature_name }}
{% endfor %}
    class_labels:
{% for class_label in class_labels %}
        - {{ class_label }}
{% endfor %}
""", trim_blocks=True)



import logging

from baseten.common import api
from baseten.common.core import raises_api_error

logger = logging.getLogger(__name__)


@raises_api_error
def add_external_connection(connection_id: str, connection_uri: str) -> dict:
    """Adds an external data source connection to the user account, which can be referred to by its connection_id in
    every workflow.

    Args:
        connection_id (str): A unique identifier for this connection. This ID will be used in the workflows
            to refer to this connection. Ex: 'prod_read_only_postgres_conn'
        connection_uri (str): Ex: 'postgresql://user:password@1.1.1.1:5432/postgresdb?optional_param=value'
    """
    return api.add_external_connection(connection_id, connection_uri)


@raises_api_error
def external_connections() -> dict:
    return api.external_connections()



import itertools
from typing import List

from baseten.workflow.atom import Atom, RegisteredAtom


def _tree_copy(tree_root, needle):
    """Perform deep copy, return new root and node corresponding to 'needle' if found."""
    new_tree_root = _Node(tree_root.atom)
    needle_counterpart = None

    if tree_root == needle:
        needle_counterpart = new_tree_root

    for c in tree_root.children:
        child_copy, child_needle_counterpart = _tree_copy(c, needle)
        new_tree_root.add_child(child_copy)
        needle_counterpart = needle_counterpart or child_needle_counterpart
    return new_tree_root, needle_counterpart


class _Node:
    """_Node is unit of graph that tracks atom flow.

    Used by Worklet internally, shouldn't be used outside.
    """

    counter = itertools.count()

    def __init__(self, atom: Atom):
        """Constructor for _Node"""
        self._children = []
        self._atom = atom
        # TODO(pankaj): This is temporary, node doen't need unique id,
        # created here for ease during publication. Should be generated
        # during publish.
        self._id = next(_Node.counter)

    @property
    def id(self):
        return self._id

    @property
    def children(self):
        return self._children

    @property
    def atom(self):
        return self._atom

    def shallow_copy(self):
        """Children are not copied deep, just references. Id is not copied and is freshly created."""
        new_node = _Node(self.atom)
        new_node._children = self._children.copy()
        return new_node

    def add_child(self, node):
        """Mutation, expected to be called only from inside Worklet."""
        self._children.append(node)
        return self


def wrap(atom):
    """Wrap an atom into a worklet. This is the primary way of creating worklets."""
    return Worklet(_Node(atom))


def wrap_with_name(atom, name):
    """Wrap an atom into a worklet, also specifying name."""
    return Worklet(_Node(atom), name)


class Worklet:
    """
    Worklet is a named rooted graph of work nodes

    Two nodes in the graph are special for Worklet, root and last.
    Worklet has only one entry point, the root node but it can have
    many leaf nodes each of which creates a separate output. Of these
    the last node is considered special, it is where any additional
    nodes get added, using e.g. the then method. Output of last node
    is considered the output of worklet.
    """

    def __init__(self, root: _Node, name: str = None, last: _Node = None):
        """
        Worklet is expected to be created mostly via wrapping atom so
        reference to _Node is ok here.
        """
        self._name = name
        self._root = root
        self._last = last
        if not last:
            self._last = root

    @property
    def root(self):
        return self._root

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def last(self):
        return self._last

    def deep_copy(self):
        new_root, new_worklet_last = _tree_copy(self._root, self._last)
        return Worklet(root=new_root, name=self._name, last=new_worklet_last)

    def then(self, worklet):
        new_worklet = self.deep_copy()
        if isinstance(worklet, list):
            for item in worklet:
                if isinstance(item, Atom):
                    new_worklet._last.add_child(_Node(item))
                elif isinstance(item, Worklet):
                    new_worklet._last.add_child(item.root)
                else:
                    raise ValueError(f'Unexpected type for list element in then: {type(item)}')
            new_worklet._last = new_worklet._last.children[0]

        elif isinstance(worklet, Atom):
            node = _Node(worklet)
            new_worklet._last.add_child(node)
            new_worklet._last = node

        elif isinstance(worklet, Worklet):
            new_worklet._last.add_child(worklet.root)
            new_worklet._last = worklet.last
        else:
            raise ValueError(f'Unexpected type for then: {type(worklet)}')

        return new_worklet

    def __rshift__(self, other):
        return self.then(other)

    def to_json(self):
        if not self._name:
            raise ValueError('A worklet should have a name')
        if not self._root:
            raise ValueError('A worklet should have work to do')

        # TODO(pankaj) node ids should be generated via topological sorting
        # this will be better because it will generate more stable ids, which
        # may be good for debugging/logging purposes.
        nodes = [_node_payload(node) for node in _all_nodes(self._root)]

        return {
            'name': self._name,
            'entry_point': self._root.id,
            'nodes': nodes,
            'exit_point': self._last.id,
        }


def _all_nodes(root: _Node) -> List[_Node]:
    nodes = []

    def accumulate_nodes(node):
        nodes.append(node)
    _walk_graph(root, accumulate_nodes, set())
    return nodes


def _walk_graph(node: _Node, func, seen):
    """Execute func on each node of graph, breaks cycles."""
    if not node:
        return

    if node in seen:
        return
    seen.add(node)

    func(node)

    for child in node.children:
        _walk_graph(child, func, seen)


def _node_payload(node: _Node):
    atom = node.atom
    if isinstance(atom, RegisteredAtom):
        return {
            'id': node.id,
            'children': [n.id for n in node.children],
            'node_name': atom.node_name,
            'atom': {
                'name': atom.name,
                'conf': atom.conf,
            },
        }

    raise ValueError('Unsupported Atom type.')



"""Console script for baseten."""
import functools
import json
import logging
import sys
from urllib.parse import urlparse

import click
import keyring

from baseten.common import api
from baseten.common.core import KEYRING_SERVICE_NAME
from baseten.common.settings import CONFIG_FILE_PATH
from baseten.common.util import setup_logger

logger = logging.getLogger(__name__)


def ensure_login(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        api_key = keyring.get_password(KEYRING_SERVICE_NAME, username=None)
        if not api_key:
            click.echo('You must first run the `baseten login` cli command.')
            sys.exit()
        result = func(*args, **kwargs)
        return result
    return wrapper


@click.group()
def cli_group():
    setup_logger('baseten', logging.INFO)


@cli_group.command()
@click.option('--server_url', prompt='BaseTen server URL')
def configure(server_url):
    if not _is_valid_url(server_url):
        click.echo('That is not a valid URL.')
        return
    with open(CONFIG_FILE_PATH, 'w') as config_file:
        config_file.write(server_url)
    click.echo('Saved server URL.')


@cli_group.command()
@click.option('--api_key', prompt='BaseTen API key', hide_input=True)
def login(api_key):
    keyring.set_password(KEYRING_SERVICE_NAME, username=None, password=api_key)
    click.echo('Saved API key.')


@cli_group.command()
@ensure_login
def models():
    user_models = api.models()
    click.echo('Your models:\n{}'.format(json.dumps(user_models, indent=4)))


@cli_group.command()
@click.option('--model_id', prompt='The model whose versions are being retrieved')
@ensure_login
def model_versions(model_id):
    user_models = api.models_versions(model_id)
    click.echo('Your model\'s versions:\n{}'.format(json.dumps(user_models, indent=4)))


def _is_valid_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except ValueError:
        return False


if __name__ == '__main__':
    cli_group()



import functools

from requests import RequestException

KEYRING_SERVICE_NAME = 'baseten'
BASETEN_CONFIG_DIRECTORY = 'baseten_model_configs/models'


class Error(Exception):
    """Base Baseten Error"""
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class AuthorizationError(Error):
    """Raised in places where the user needs to be logged in and is not."""
    pass


class FrameworkNotSupportedError(Error):
    """Raised in places where the user attempts to use Baseten with an unsupported framework"""
    pass


class ModelNotSupportedError(Error):
    """Raised when a user attempts to use unsupported models for a
    framework that is otherwise supported"""
    pass


class ModelInputShapeError(Error):
    """Raised when deploying a model whose input shape doesn't match the number of feature names given by the user."""
    pass


class SampleDataInputShapeError(Error):
    """Raised when deploying a model whose input shape doesn't match the number of feature names given by the user."""
    pass


class WorkflowError(Error):
    """Base error for workflow-related errors."""
    pass


class InvokingUnpublishedWorkflowError(WorkflowError):
    """Raised when invoking a workflow's worklets before the workflow is published."""
    pass


class ExplainerNotSupportedError(Error):
    """Raised when the user attempts to deploy a configuration that doesn't make sense for the explainer"""
    pass


class ExplainerInvalidConfigrationError(Error):
    """Raised when attempting to deploy an explainer without the correct configuration required."""


class ApiError(Error):
    """Errors in calling the Baseten API."""
    def __init__(self, message, response=None):
        super().__init__(message)
        self.response = response

    def __str__(self):
        error_str = self.message
        if self.response is not None:  # non-200 Response objects are falsy, hence the not None.
            error_str = f'{error_str}\n<Server response: {self.response.content}>'
        return error_str


def raises_api_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestException as err:
            raise ApiError(f'Error calling {func.__name__}.', response=err.response) from err
    return wrapper



import os
from pathlib import Path

SKLEARN = 'sklearn'
TENSORFLOW = 'tensorflow'

DEBUG = False
CONFIG_FILE_PATH = f'{Path.home()}/.baseten_config'


def _api_url_base() -> str:
    try:
        with open(CONFIG_FILE_PATH) as config_file:
            server_url = config_file.read().strip().rstrip('/')
            if server_url:
                return server_url
    except IOError:
        pass
    baseten_cli_env = os.environ.get('BASETEN_ENV')
    if baseten_cli_env == 'production':
        return 'https://app.baseten.co'
    if baseten_cli_env == 'staging':
        return 'https://staging.app.baseten.co'
    return 'http://127.0.0.1:8000'


API_URL_BASE = _api_url_base()



import inspect
from typing import Callable, List

import libcst as cst


class LambdaCollector(cst.CSTVisitor):
    def __init__(self):
        self.lambdas = []

    def visit_Lambda(self, node: cst.Lambda):
        self.lambdas.append(node)


def _code_obj_similar(c1, c2):
    if c1.co_argcount != c2.co_argcount:
        return False
    if c1.co_cellvars != c2.co_cellvars:
        return False
    if c1.co_code != c2.co_code:
        return False
    if c1.co_consts != c2.co_consts:
        return False
    if c1.co_freevars != c2.co_freevars:
        return False
    if c1.co_kwonlyargcount != c2.co_kwonlyargcount:
        return False
    if c1.co_names != c2.co_names:
        return False
    return True


def _gen_lambda_code_obj(cst_module, cst_lambda: cst.Lambda):
    code = cst_module.code_for_node(cst_lambda)
    outer_module_code_obj = compile(code, '', 'single')
    lambda_code_obj = outer_module_code_obj.co_consts[0]
    return lambda_code_obj


def _find_matching_lambda(lambda_code_obj_needle,
                          lambda_code_objs_in_src_line: List[cst.Lambda],
                          cst_module) -> cst.Lambda:
    for cst_lambda in lambda_code_objs_in_src_line:
        lambda_code_obj = _gen_lambda_code_obj(cst_module, cst_lambda)
        if _code_obj_similar(lambda_code_obj, lambda_code_obj_needle):
            return cst_lambda
    return None


def get_lambda_source(lambda_callable: Callable):
    code_obj = lambda_callable.__code__
    source_line = inspect.getsourcelines(lambda_callable)[0][0].lstrip()
    cst_module = cst.parse_module(source_line)
    visitor = LambdaCollector()
    cst_module.visit(visitor)
    lambdas = visitor.lambdas
    matching_lambda = _find_matching_lambda(code_obj, lambdas, cst_module)
    if not matching_lambda:
        raise Exception('Unable to find matching lambda, this should not happen, please report a bug.')
    return cst_module.code_for_node(matching_lambda)



from abc import ABC
from typing import List, Union

import numpy as np
import pandas as pd

from baseten.common.core import (ExplainerInvalidConfigrationError,
                                 ExplainerNotSupportedError)

FRAMEWORK_TO_SHAP_TREE_MODELS_MAP = {
    'sklearn': {
        'RandomForestRegressor',
        'IsolationForest',
        'ExtraTreesRegressor',
        'DecisionTreeRegressor',
        'DecisionTreeClassifier',
        'RandomForestClassifier',
        'ExtraTreesClassifier',
        'GradientBoostingRegressor',
        'MeanEstimator',
        'QuantileEstimator',
        'DummyRegressor',
        'GradientBoostingClassifier',
        'LogOddsEstimator',
        'DummyClassifier'
    }
}

FRAMEWORK_TO_SHAP_LINEAR_MODELS_MAP = {
    'sklearn': {
        'LinearRegression',
        'Ridge',
        'RidgeClassifier',
        'RidgeCV',
        'Lasso',
        'MultiTaskLasso',
        'ElasticNet',
        'ElasticNetCV',
        'MultiTaskElasticNet',
        'LassoLars',
        'OrthogonalMatchingPursuit',
        'OrthogonalMatchingPursuitCV',
        'BayesianRidge',
        'ARDRegression',
        'LogisticRegression',
        'SGDClassifier',
        'SGDRegressor',
        'Perceptron',
        'PassiveAggressiveClassifier',
        'PassiveAggressiveRegressor',
        'TheilSenRegressor',
        'RANSACRegressor',
        'HuberRegressor',
    }
}

BLACKBOX_FRAMEWORKS_MODELS_MAP = {
    'sklearn': {
        'RandomForestRegressor',
        'IsolationForest',
        'ExtraTreesRegressor',
        'DecisionTreeRegressor',
        'DecisionTreeClassifier',
        'RandomForestClassifier',
        'ExtraTreesClassifier',
        'GradientBoostingRegressor',
        'MeanEstimator',
        'QuantileEstimator',
        'DummyRegressor',
        'GradientBoostingClassifier',
        'LogOddsEstimator',
        'DummyClassifier',
        'LinearRegression',
        'Ridge',
        'RidgeClassifier',
        'RidgeCV',
        'Lasso',
        'MultiTaskLasso',
        'ElasticNet',
        'ElasticNetCV',
        'MultiTaskElasticNet',
        'LassoLars',
        'OrthogonalMatchingPursuit',
        'OrthogonalMatchingPursuitCV',
        'BayesianRidge',
        'ARDRegression',
        'LogisticRegression',
        'SGDClassifier',
        'SGDRegressor',
        'Perceptron',
        'PassiveAggressiveClassifier',
        'PassiveAggressiveRegressor',
        'TheilSenRegressor',
        'RANSACRegressor',
        'HuberRegressor',
    }
}


class BaseModelExplainer(ABC):
    name = 'BaseExplainer'
    requires_background_data = False
    requires_features_names = False


class TreeExplainer(BaseModelExplainer):
    name = 'ShapTreeExplainer'
    requires_background_data = False
    description = 'Uses Tree SHAP algorithms to explain the output of ensemble tree models.'
    framework_model_support = FRAMEWORK_TO_SHAP_TREE_MODELS_MAP


class LinearExplainer(BaseModelExplainer):
    name = 'ShapLinearExplainer'
    requires_background_data = True
    description = 'Computes SHAP values for a linear model.'
    framework_model_support = FRAMEWORK_TO_SHAP_LINEAR_MODELS_MAP


class KernelExplainer(BaseModelExplainer):
    name = 'ShapKernelExplainer'
    requires_background_data = True
    description = 'Uses the Kernel SHAP method to explain the output of any function.'
    framework_model_support = BLACKBOX_FRAMEWORKS_MODELS_MAP


class AnchorTabular(BaseModelExplainer):
    name = 'AlibiAnchorTabular'
    requires_background_data = True
    requires_features_names = True
    description = 'Uses the Alibi Anchor method to explain the output of tabular data.'
    framework_model_support = BLACKBOX_FRAMEWORKS_MODELS_MAP


class AnchorCounterFactual(BaseModelExplainer):
    name = 'AlibiCounterFactual'
    requires_background_data = False
    description = 'Uses the Alibi Anchor method to explain the output of tabular data.'
    framework_model_support = BLACKBOX_FRAMEWORKS_MODELS_MAP


NAME_TO_EXPLAINER_MAP = {
    'BaseModelExplainer': BaseModelExplainer,
    'ShapTreeExplainer': TreeExplainer,
    'ShapLinearExplainer': LinearExplainer,
    'ShapKernelExplainer': KernelExplainer,
    'AlibiAnchorTabular': AnchorTabular,
    'AlibiCounterFactual': AnchorCounterFactual,
}


def raises_invalid_explainer_configuration(
    explainer_name: str,
    model_framework: str,
    model_type: str,
    feature_data: Union[np.ndarray, pd.DataFrame, List] = None,
    feature_names=None,
) -> None:
    if explainer_name not in NAME_TO_EXPLAINER_MAP:
        raise ExplainerNotSupportedError(f'Explainer: {explainer_name} is not known to BaseTen.')
    elif model_type not in NAME_TO_EXPLAINER_MAP[explainer_name].framework_model_support.get(model_framework, {}):
        raise ExplainerNotSupportedError(
            f'Explainer: {explainer_name} is not compatible for {model_framework}:{model_type}.')
    elif NAME_TO_EXPLAINER_MAP[explainer_name].requires_background_data and feature_data is None:
        raise ExplainerInvalidConfigrationError(
            f'Explainer: {explainer_name} requires background data for explanation.')
    elif NAME_TO_EXPLAINER_MAP[explainer_name].requires_features_names and feature_names is None:
        raise ExplainerInvalidConfigrationError(f'Explainer: {explainer_name} requires feature names for explanation.')



import base64
import copy
import json
import logging
import os
from typing import List, Union

import numpy as np
import pandas as pd
from colorama import Fore, Style

from baseten.common import core
from baseten.templates.transform_template import (
    feature_transform_config_template, feature_transform_python_template,
    output_transform_config_template, output_transform_python_template,
    transform_view_template)

TRANSFORM_INPUT_TYPE_TO_PYTHON_TYPE = {
    'categorical': 'str',
    'float': 'float',
    'int': 'int',
    'text': 'str',
    'image': 'bytes',
    'url': 'str'
}

# list from https://scikit-learn.org/stable/developers/advanced_installation.html
SKLEARN_REQ_MODULE_NAME = {
    'numpy',
    'scipy',
    'joblib',
    'scikit-learn',
    'threadpoolctl',
}

# list from https://www.tensorflow.org/install/pip
# if problematic, lets look to https://www.tensorflow.org/install/source
TENSORFLOW_REQ_MODULE_NAME = {
    'tensorflow',
}


LOG_COLORS = {
    logging.ERROR: Fore.RED,
    logging.DEBUG: Fore.MAGENTA,
    logging.WARNING: Fore.YELLOW,
    logging.INFO: Fore.GREEN,
}


class ColorFormatter(logging.Formatter):
    def format(self, record, *args, **kwargs):
        new_record = copy.copy(record)
        if new_record.levelno in LOG_COLORS:
            new_record.levelname = "{color_begin}{level}{color_end}".format(
                level=new_record.levelname,
                color_begin=LOG_COLORS[new_record.levelno],
                color_end=Style.RESET_ALL,
            )
        return super(ColorFormatter, self).format(new_record, *args, **kwargs)


def setup_logger(package_name, level):
    baseten_logger = logging.getLogger(package_name)
    baseten_logger.setLevel(level)
    handler = logging.StreamHandler()
    formatter = ColorFormatter(fmt='%(levelname)s %(message)s')
    handler.setFormatter(formatter)
    if baseten_logger.hasHandlers():
        baseten_logger.handlers.clear()
    baseten_logger.addHandler(handler)


def generate_feature_transform_template(model_version_id: str,
                                        transform_name: str,
                                        input_names: List[str],
                                        input_types: List[str]):
    """Creates boilerplate code and config for the transform.

    Args:
        model_version_id(str)
        transform_name (str): Name given to the new transform; e.g 'my_awesome_feature_transform'
        input_names (List[str]): The names of the transform's input keys
        input_types (List[str]): The types of the transform's input keys (must be
            in TRANSFORM_INPUT_TYPE_TO_PYTHON_TYPE)
    """
    if not input_names:
        raise ValueError('Input names and input types must be 1 to 1')
    if input_types and (len(input_names) != len(input_types)):
        raise ValueError('Input names and input types must be 1 to 1')
    elif input_names and not input_types:
        # TODO attempt to infer types somewhere else (eg from training data)
        input_types = ['float' for _ in input_names]
    if not _validate_input_types(input_types):
        raise ValueError(f'Types must be one of {list(TRANSFORM_INPUT_TYPE_TO_PYTHON_TYPE.keys())}')
    input_data = [(x, y) for x, y in zip(input_names, input_types)]
    data = {
        'transform_name': transform_name,
        'input_data': input_data
    }

    path = _config_path_for_model_version(model_version_id)
    with open(f'{path}{transform_name}_config.yml', 'w') as f:
        f.writelines(feature_transform_config_template.render(data))
    with open(f'{path}{transform_name}_view.yml', 'w') as f:
        f.writelines(transform_view_template.render(data))
    with open(f'{path}{transform_name}.py', 'w') as f:
        f.writelines(feature_transform_python_template.render({'input_data': input_data}))
    return path


def generate_output_transform_template(model_version_id: str, transform_name: str):
    """Creates boilerplate code and config for the output transform.

    Args:
        model_version_id(str)
        transform_name (str): Name given to the new transform; e.g 'my_awesome_model_output_transform'
    """
    path = _config_path_for_model_version(model_version_id)
    with open(f'{path}{transform_name}_config.yml', 'w') as f:
        f.writelines(output_transform_config_template.render({'transform_name': transform_name}))
    with open(f'{path}{transform_name}.py', 'w') as f:
        f.writelines(output_transform_python_template.render())
    return path


def coerce_data_as_numpy_array(data: Union[np.ndarray, pd.DataFrame, List]) -> np.ndarray:
    """Validates that data can be coerced into a numpy array

    Args:
        data (Union[np.ndarray, pd.DataFrame, List]): The data to be transformed.

    Raises:
        TypeError: If data is wrong type.

    Returns:
        np.ndarray: A numpy array of data.
    """
    if not isinstance(data, (np.ndarray, pd.DataFrame, list)):
        raise TypeError(f'Data must be one of type [np.ndarray, pd.DataFrame, list], got {type(data)}')
    return np.array(data)


def base64_encoded_json_str(obj):
    return base64.b64encode(str.encode(json.dumps(obj))).decode('utf-8')


def _config_path_for_model_version(model_version_id: str) -> str:
    path = f'{core.BASETEN_CONFIG_DIRECTORY}/{model_version_id}/'
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def _validate_input_types(input_types):
    return all([input_type in TRANSFORM_INPUT_TYPE_TO_PYTHON_TYPE.keys() for input_type in input_types])


def zipdir(path, zip_handler):
    for root, dirs, files in os.walk(path):
        relative_root = ''.join(root.split(path))
        for _file in files:
            zip_handler.write(
                os.path.join(root, _file), os.path.join(f'model{relative_root}', _file))


def print_error_response(response):
    print(Fore.YELLOW + f'{response["message"]}')
    print('---------------------------------------------------------------------------')
    print(Fore.GREEN + 'Stack Trace:')
    if 'exception' in response:
        excp = response['exception']
        for excp_st_line in excp['stack_trace']:
            print(Fore.GREEN + '---->' + Fore.WHITE + f'{excp_st_line}')
        print(Fore.RED + f'Exception: {excp["message"]}')


def pip_freeze():
    """
    This spawns a subprocess to do a pip freeze programmatically. pip is generally not supported as an API or threadsafe

    Returns: The result of a `pip freeze`

    """
    stream = os.popen('pip freeze -qq')
    this_env_requirements = [line.strip() for line in stream.readlines()]

    return this_env_requirements


def _get_entries_for_packages(list_of_requirements, desired_requirements):
    name_to_req_str = {}
    for req_name in desired_requirements:
        for full_req_str in list_of_requirements:
            if req_name == full_req_str.split('==')[0]:
                name_to_req_str[req_name] = full_req_str
    return name_to_req_str


def infer_sklearn_packages():
    return _get_entries_for_packages(pip_freeze(), SKLEARN_REQ_MODULE_NAME)


def infer_tensorflow_packages():
    return _get_entries_for_packages(pip_freeze(), TENSORFLOW_REQ_MODULE_NAME)



import logging
import os
import tempfile
import zipfile

import requests
import tensorflow as tf

import baseten
from baseten.baseten_deployed_model import BasetenDeployedModel

MNIST_PRETRAINED_URL = 'https://baseten-public.s3-us-west-2.amazonaws.com/models/mnist.zip'
logger = logging.getLogger(__name__)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'


def build_model() -> tf.keras.Model:
    """Constructs the ML model used to predict handwritten digits."""

    image = tf.keras.layers.Input(shape=(28, 28, 1))

    y = tf.keras.layers.Conv2D(filters=32,
                               kernel_size=5,
                               padding='same',
                               activation='relu')(image)
    y = tf.keras.layers.MaxPooling2D(pool_size=(2, 2),
                                     strides=(2, 2),
                                     padding='same')(y)
    y = tf.keras.layers.Conv2D(filters=32,
                               kernel_size=5,
                               padding='same',
                               activation='relu')(y)
    y = tf.keras.layers.MaxPooling2D(pool_size=(2, 2),
                                     strides=(2, 2),
                                     padding='same')(y)
    y = tf.keras.layers.Flatten()(y)
    y = tf.keras.layers.Dense(1024, activation='relu')(y)
    y = tf.keras.layers.Dropout(0.4)(y)

    probs = tf.keras.layers.Dense(10, activation='softmax')(y)

    model = tf.keras.models.Model(inputs=image, outputs=probs, name='mnist')

    return model


def scale_inputs(input_images):
    return input_images / 255.


def get_test_data():
    _, (X_test, Y_test) = get_training_data()
    return X_test, Y_test


def get_training_data():
    (train_images, train_labels), (test_images, test_labels) = tf.keras.datasets.mnist.load_data()
    train_images = train_images.reshape((-1, 28, 28, 1))
    test_images = test_images.reshape((-1, 28, 28, 1))
    train_images = scale_inputs(train_images)
    test_images = scale_inputs(test_images)
    return (train_images, train_labels), (test_images, test_labels)


def train_mnist(train_epochs=10) -> tf.keras.Model:
    """Trains the MNIST tensorflow model from scratch.

    Args:
        train_epochs (int, optional): Number of epochs to train for. Defaults to 10.

    Returns:
        tf.keras.Model: A trained model.
    """
    (train_images, train_labels), (test_images, test_labels) = get_training_data()
    model = build_model()
    model.compile(loss=tf.keras.losses.sparse_categorical_crossentropy,
                  optimizer=tf.keras.optimizers.Adadelta(),
                  metrics=['accuracy', 'sparse_categorical_crossentropy'])
    model.fit(train_images,
              train_labels,
              epochs=train_epochs,
              validation_data=(test_images, test_labels))

    return model


def deploy_pretrained_model() -> BasetenDeployedModel:
    """Fetches and deploys a pretrained MNIST handwritten digit classifier.

    Returns:
        BasetenDeployedModel: A deployed BaseTen model.
    """
    logger.info('Fetching MNIST model data from s3.')
    model_request = requests.get(MNIST_PRETRAINED_URL)
    logger.info('Fetched MNIST model data from s3.')
    filename = 'model.zip'
    temp_dir = tempfile.TemporaryDirectory()
    model_filepath = os.path.join(temp_dir.name, filename)
    with open(model_filepath, 'wb') as f:
        f.write(model_request.content)
    logger.info('Extracting model contents.')
    with zipfile.ZipFile(model_filepath, 'r') as zip_file:
        zip_file.extractall(temp_dir.name)
    logger.info('Restoring model.')
    tf_model = tf.keras.models.load_model(temp_dir.name)
    logger.info('Deploying model to BaseTen.')
    model = baseten.deploy(tf_model, 'mnist-tensorflow')
    logger.info('Deployed model to BaseTen.')
    return model



import importlib

BASETEN_PRETRAINED_MODELS = {  # Map model names to files where they are defined.
    'mnist_tensorflow': 'mnist'
}


def deploy_pretrained_model(model_name: str):
    if model_name not in BASETEN_PRETRAINED_MODELS:
        raise Exception(f'{model_name} is not available as a pretrained baseten model.')
    model_file = BASETEN_PRETRAINED_MODELS[model_name]
    module = importlib.import_module(f'baseten.examples.{model_file}')
    return module.deploy_pretrained_model()



# flake8: noqa

from jinja2 import Template

feature_transform_python_template = Template("""
def transform_input(input_data):
    \"\"\"
    Transforms input dict of feature data into the model input as given by the input signature of the model.

    Args:
        input_data (Dict): a dict of input names to values
        {
{% for input_name, _ in input_data %}
                '{{ input_name }}': INPUT_VALUE,
{% endfor %}

        }
     Returns:
        A list of transformed values.
    \"\"\"
    # TODO: Your transform code goes here:
    return []
""", trim_blocks=True)

output_transform_python_template = Template("""
def transform_model_output(model_output_data):
    \"\"\"
    Transforms the output of the model into a human-understandable string.

    Args:
        model_output_data (Dict): The output of the model as a dict with keys:
            'prediction' (Union[float, List]): float or list depending on the model
            'probabilities' (List[float], optional)
            'explainability' (Dict, optional)
     Returns:
        A human-understandable string.
    \"\"\"
    # TODO: Your model output transform code goes here:
    return ''
""", trim_blocks=True)

feature_transform_config_template = Template(
"""transform_name: {{ transform_name }}
transform_type: feature_transform
transform_inputs:
{% for input_name, input_type in input_data %}
    - {{ input_name }}:
        type: {{ input_type }}
{% endfor %}""", trim_blocks=True)

output_transform_config_template = Template(
"""transform_name: {{ transform_name }}
transform_type: output_transform
""", trim_blocks=True)


transform_view_template = Template(
"""transform_name: {{ transform_name }}
transform_inputs:
{% for input_name, input_type in input_data %}
    - {{ input_name }}:
        display_name: {{ input_name }}
        type: {{ input_type }}
{% if input_type == 'categorical' %}
        categories: []
{% endif %}
{% if input_type == 'float' %}
        min: # optional minimum for float value
        max: # optional maximum for float value
{% endif %}
{% if input_type == 'int' %}
        min: # optional minimum for int value
        max: # optional maximum for int value
{% endif %}
{% endfor %}""", trim_blocks=True)



class Query:
    def __init__(self, name: str, query_str: str, external_connection: str = None):
        self._name = name
        self._query_str = query_str
        self._external_connection = external_connection

    def to_json(self):
        return {
            'name': self._name,
            'query_str': self._query_str,
            'external_connection': self._external_connection,
        }



import logging

import yaml

from baseten.common import api
from baseten.common.core import raises_api_error

logger = logging.getLogger(__name__)


class BasetenWorkflow:
    """A workflow. Provide either the workflow_id or the workflow_version_id."""
    def __init__(self,
                 workflow_id: str,
                 workflow_version_id: str = None,
                 workflow_name: str = None):
        if not workflow_id and not workflow_version_id:
            raise ValueError(
                'Either workflow_id or workflow_version_id must be provided.')
        self._workflow_id = workflow_id
        self._workflow_version_id = workflow_version_id
        self._workflow_name = workflow_name

    @property
    def workflow_name(self):
        return self._workflow_name

    @property
    def workflow_version_id(self):
        return self._workflow_version_id

    @raises_api_error
    def create_workflow_action(self, action_name: str, action_code_path: str,
                               action_entrypoint: str):
        """Create an action.
        Args:
            action_name (str): The name of the action
            action_code_path (str): The path to the action's .py file.
            action_entrypoint (str): If action_code_path is a .py file, the name of the
                function which is the entrypoint for the action.
        Returns:
            [type]: [description]
        """
        config = {'entrypoint': action_entrypoint}
        action_code = open(action_code_path, 'r').read()
        action_data = {'code': action_code, 'config': config}
        return api.create_workflow_action(self._workflow_id, action_name, action_data)

    @raises_api_error
    def publish(self):
        """Publish the workflow.
        """
        return api.publish_workflow(self._workflow_version_id)

    @raises_api_error
    def update_workflow(self, workflow_yaml_filepath):
        """Update a workflow.
        Args:
            workflow_yaml_filepath (str): The workflow yaml file path.
        """
        workflow_config = yaml.safe_load(open(workflow_yaml_filepath, 'r'))
        resp = api.update_workflow(self._workflow_id, workflow_config)
        # Setting the version ID so that a future `publish` call publishes this new workflow version.
        self._workflow_version_id = resp['id']
        return resp



import base64
import inspect
from typing import Callable

import cloudpickle

from baseten.workflow.atom import RegisteredAtom
from baseten.workflow.lambda_source import get_lambda_source


class SendSlackMessage(RegisteredAtom):
    def __init__(self, webhook_url: str, format: str, task_name: str = 'Send Slack Message'):
        super().__init__('baseten.send_slack_message',
                         task_name,
                         {
                             'webhook_url': webhook_url,
                             'format': format,
                         })


class InvokeModel(RegisteredAtom):
    def __init__(self, model_id: str, task_name: str = 'Invoke Model'):
        super().__init__('baseten.invoke_model',
                         task_name,
                         {
                             'model_id': model_id,
                         })


class ExecuteQuery(RegisteredAtom):
    def __init__(self, query_name: str, task_name: str = 'Execute Query'):
        super().__init__('baseten.execute_query',
                         task_name,
                         {
                             'query_name': query_name,
                         })


def _extract_params(python_callable: Callable):
    def annotation_to_string(annotation):
        if hasattr(annotation, '__name__'):
            return annotation.__name__
        return str(annotation)

    def param_dict(param: inspect.Parameter):
        default = param.default if param.default != inspect.Signature.empty else None
        return {
            'name': param.name,
            'default': default,
            'annotation': annotation_to_string(param.annotation),
            'kind': str(param.kind),
        }
    return [param_dict(p) for p in inspect.Signature.from_callable(python_callable).parameters.values()]


def _get_source_lines(python_callable: Callable):
    # TODO(pankaj) Tackle lambda code better, currently it picks up the entire source code line
    if inspect.isfunction(python_callable):
        if python_callable.__name__ == '<lambda>':
            return [get_lambda_source(python_callable)]
        return inspect.getsourcelines(python_callable)[0]
    return inspect.getsourcelines(type(python_callable))[0]


def _ser_obj(obj):
    pickled = cloudpickle.dumps(obj)
    return base64.b64encode(pickled).decode('utf-8')


def _extract_callable_properties(python_callable: Callable):
    name = python_callable.__name__ if hasattr(python_callable, '__name__') else ''
    args_metadata = _extract_params(python_callable)
    source = [line[:-1] if line and line[-1] == '\n' else line
              for line in _get_source_lines(python_callable)]
    ser_code = _ser_obj(python_callable)
    return name, args_metadata, source, ser_code


class Py(RegisteredAtom):
    """Generate Atom from python callable."""
    _ATOM_NAME = 'baseten.py'

    def __init__(self,
                 python_callable: Callable,
                 expand_input_dict=True,
                 task_name='Execute Python Callable',
                 doc=None):
        if not callable(python_callable):
            raise ValueError('supplied python_callable is not a callable.')

        c_name, c_args_metadata, c_source, c_ser_code = _extract_callable_properties(python_callable)
        # TODO(pankaj) extract doc from callable python docstring
        doc = doc or ''
        conf = {
            'callable_name': c_name,
            'callable_args_metadata': c_args_metadata,
            'callable_source': c_source,
            'callable_encoded_serialized_code': c_ser_code,
            'callable_expand_input_dict': expand_input_dict,
            'doc': doc,
        }
        super().__init__(Py._ATOM_NAME, task_name, conf)


class Map(RegisteredAtom):
    """Transform input.

    Convenience method to ignore the context and transform just he input. Allows
    for lean inline transformation chaining.

    Example:
        Map(lambda _: 10) >> Map(lambda x: x * 2) >> Map(lambda x: print(x))
        Prints 20
    """

    _ATOM_NAME = 'baseten.map'

    def __init__(self,
                 python_callable: Callable,
                 expand_input_dict=False,
                 task_name='Map',
                 doc=None):
        if not callable(python_callable):
            raise ValueError('supplied python_callable is not a callable.')

        c_name, c_args_metadata, c_source, c_ser_code = _extract_callable_properties(python_callable)
        # TODO(pankaj) extract doc from callable python docstring
        doc = doc or ''
        conf = {
            'callable_name': c_name,
            'callable_args_metadata': c_args_metadata,
            'callable_source': c_source,
            'callable_encoded_serialized_code': c_ser_code,
            'callable_expand_input_dict': expand_input_dict,
            'doc': doc,
        }
        super().__init__(Map._ATOM_NAME, task_name, conf)



import json
import logging
import os
import pathlib
import tempfile
import time
from typing import Dict, List, Union

import h5py
import numpy as np
import pandas as pd
import requests
import yaml

from baseten.common import api
from baseten.common.core import (BASETEN_CONFIG_DIRECTORY,
                                 SampleDataInputShapeError, raises_api_error)
from baseten.common.explainer_config import \
    raises_invalid_explainer_configuration
from baseten.common.util import (coerce_data_as_numpy_array,
                                 generate_feature_transform_template,
                                 generate_output_transform_template)
from baseten.templates.model_template import model_template
from baseten.workflows import BasetenWorkflow

logger = logging.getLogger(__name__)

REQUIREMENTS_INSTALLATION_STATUS_RETRY_INTERVAL_SEC = 3
REQUIREMENTS_INSTALLATION_STATUS_MAX_TRIES = 20


def _build_h5_data_object(feature_data_np, target_data, metadata, data_temp_directory):
    data_temp_file = pathlib.PurePath(data_temp_directory, 'tmp.h5')
    data = {'features': feature_data_np}
    if np.any(target_data):
        data['targets'] = coerce_data_as_numpy_array(target_data)
    if np.any(metadata):
        data['metadata'] = json.dumps(metadata)

    h5_data = h5py.File(data_temp_file, mode='w')
    h5_sample_data_group = h5_data.create_group('sample_data')
    for key, np_obj in data.items():
        if key == 'metadata':
            dt = h5py.string_dtype(encoding='utf-8')
            h5_sample_data_group.create_dataset(key, data=np_obj, dtype=dt)
        else:
            h5_sample_data_group.create_dataset(key, data=np_obj)
    h5_data.close()
    return data_temp_file


@raises_api_error
def install_requirements(req_filepath: str):
    with open(req_filepath, 'r') as fp:
        requirements_txt = fp.read()
    logger.info(f'🚀 Sending requirements to BaseTen 🚀')
    resp = api.install_requirements(requirements_txt)
    status = resp['status']
    if status == 'PROCESSING':
        logger.info('🐳 Requirements are being installed 🐳')

        requirement_id = resp['id']
        tries = 0
        while tries < REQUIREMENTS_INSTALLATION_STATUS_MAX_TRIES:
            time.sleep(REQUIREMENTS_INSTALLATION_STATUS_RETRY_INTERVAL_SEC)
            resp = api.requirement_status(requirement_id)
            if resp['status'] != 'PROCESSING':
                break
            tries += 1
        else:
            logger.info('⌛ Requirements are still being installed. Check the status by running '
                        f'baseten.requirements_status(\'{requirement_id}\') ⌛')
    if status == 'SUCCEEDED':
        logger.info('🖖 Installed requirements successfully 🖖')
    elif status == 'FAILED':
        error_message = resp['error_message']
        logger.info(f'⚠️ Failed to install requirements. Error: "{error_message}" ⚠️')


@raises_api_error
def requirements_status(requirement_id: str):
    return api.requirement_status(requirement_id)


class BasetenDeployedExplainer:
    def __init__(self, explainer_version_id: str):
        self._explainer_version_id = explainer_version_id

    @raises_api_error
    def explain(self,
                inputs: Union[List, pd.DataFrame, np.ndarray],
                metadata: Union[pd.DataFrame, List[Dict]] = None) -> List[List]:
        if isinstance(inputs, pd.DataFrame):
            inputs_list = inputs.to_dict('records')
        elif isinstance(inputs, (list, np.ndarray)):
            inputs_np_array = np.array(inputs)
            inputs_list = inputs_np_array.tolist()
        else:
            raise TypeError('predict can be called with either a list, a pandas DataFrame, or a numpy array.')

        if isinstance(metadata, pd.DataFrame):
            metadata = metadata.to_dict(orient='records')

        return api.explain_for_version(self._explainer_version_id, inputs_list, metadata)


class BasetenDeployedModel:
    """A model backed by baseten serving. Provide either the model_id or the model_version_id."""
    def __init__(self, model_id: str = None,
                 model_version_id: str = None,
                 model_name: str = None):
        if not model_id and not model_version_id:
            raise ValueError('Either model_id or model_version_id must be provided.')

        if model_id and model_version_id:
            raise ValueError('Must provide either model_id or model_version_id; not both.')

        self._model_id = model_id
        self._model_version_id = model_version_id
        self._model_name = model_name
        self._model_config = None

    @property
    def model_version_id(self):
        return self._model_version_id

    @raises_api_error
    def predict(self,
                inputs: Union[List, pd.DataFrame, np.ndarray],
                metadata: Union[pd.DataFrame, List[Dict]] = None) -> List[List]:
        """Invokes the model given the input dataframe.

        Args:
            inputs: The data representing one or more inputs to call the model with.
                Accepted types are: list, pandas.DataFrame, and numpy.ndarray
            metadata (Union[pd.DataFrame, List[Dict]]): Metadata key/value pairs (e.g. name, url), one for each input.

        Returns:
            A list of inferences for each given input; e.g.: [[3], [9]] would indicate the prediction for the
            first input in inputs_df is [3], and the prediction for the second is [9].

        Raises:
            TypeError: If the provided inputs is not of a supported type.
            ApiError: If there was an error communicating with the server.
        """
        if isinstance(inputs, pd.DataFrame):
            inputs_list = inputs.to_dict('records')
        elif isinstance(inputs, (list, np.ndarray)):
            inputs_np_array = np.array(inputs)
            inputs_list = inputs_np_array.tolist()
        else:
            raise TypeError('predict can be called with either a list, a pandas DataFrame, or a numpy array.')

        if isinstance(metadata, pd.DataFrame):
            metadata = metadata.to_dict(orient='records')

        if self._model_version_id:
            return api.predict_for_model_version(self._model_version_id, inputs_list, metadata)
        return api.predict_for_model(self._model_id, inputs_list, metadata)

    @raises_api_error
    def update_model_config(self, model_config_file_path: str):
        """Update the model's feature names and output class labels (if any) based on the config
        found at `model_config_file_path`

        Args:
            model_config_file_path (str): The path to the model config file
        """
        config_yaml = yaml.safe_load(open(model_config_file_path, 'r'))
        feature_names = list(config_yaml['model_features']['features'])
        class_labels = config_yaml.get('model_class_labels', [])
        api.update_model_config(self._model_version_id, feature_names, class_labels)

    @raises_api_error
    def create_data_transform(self, model_config_path: str, transform_name: str):
        """Create a new data transform.

        Args:
            model_config_path (str): The path of this model's config folder
            transform_name (str)
        """
        transform_python_file_path = f'{model_config_path}/{transform_name}.py'
        transform_config_file_path = f'{model_config_path}/{transform_name}_config.yml'
        transform_view_file_path = f'{model_config_path}/{transform_name}_view.yml'
        transform_code = open(transform_python_file_path, 'r').read()
        transform_config = yaml.safe_load(open(transform_config_file_path, 'r'))
        if os.path.exists(transform_view_file_path):
            view_config = yaml.safe_load(open(transform_view_file_path, 'r'))
        else:
            view_config = {}
        transformation_data = {'code': transform_code, 'config': transform_config}
        # TODO validations
        return api.create_data_transform(self._model_version_id, transformation_data, view_config)

    @raises_api_error
    def create_workflow(self, workflow_name: str, workflow_yaml_filepath: str) -> BasetenWorkflow:
        """Create a workflow.
        Args:
            workflow_name (str): The name of the new workflow.
            workflow_yaml_filepath (str): The workflow yaml file path.

        Returns:
            BasetenWorkflow: A BasetenWorkflow instance.
        """
        with open(workflow_yaml_filepath, 'r') as config_file:
            workflow_config = config_file.read()
        response = api.create_workflow(self._model_version_id, workflow_name, workflow_config)
        workflow = BasetenWorkflow(
            workflow_id=response['id'],
            workflow_version_id=response['workflow_versions'][0]['id'],  # The workflow version ID that was just created
            workflow_name=response['name'])
        return workflow

    @raises_api_error
    def transform_data(self, data_transform_id: str, input_data: Dict) -> Dict:
        """Invokes the transform and returns a dict with transformed data and some metadata.

        Args:
            data_transform_id (str)
            input_data (Dict):  Data to be transformed.

        Returns:
            A dict with keys: 'success' (required) and 'transform_invocation_id', 'transform_output', 'error' (optional)
        """
        return api.transform_data(data_transform_id, input_data)

    def generate_feature_transform_template(self, transform_name: str, input_names: List[str], input_types: List[str]):
        return generate_feature_transform_template(model_version_id=self.model_version_id,
                                                   transform_name=transform_name,
                                                   input_names=input_names,
                                                   input_types=input_types)

    def generate_output_transform_template(self, transform_name: str):
        return generate_output_transform_template(model_version_id=self.model_version_id,
                                                  transform_name=transform_name)

    @raises_api_error
    def update_transform_view(self, data_transform_id: str, transform_view_config_file_path: str):
        """Update the transform's view config based on the config file found at `transform_view_file_path`

        Args:
            data_transform_id (str)
            transform_view_config_file_path (str): The path to the transform view config
                file. e.g: './baseten_model_configs/models/2vq0y3d/my_awesome_transform_view.yml'
        """
        view_config = yaml.safe_load(open(transform_view_config_file_path, 'r'))
        return api.update_transform_view(data_transform_id, view_config)

    @raises_api_error
    def set_primary(self):
        """Promote this version of the model as the primary version.
        Raises:
            ApiError: If there was an error communicating with the server.
        """
        if not self._model_version_id:
            raise ValueError('Only a BasetenDeployedModel backed by a model_version can be set as primary.')
        return api.set_primary(self._model_version_id)

    @raises_api_error
    def upload_sample_data(self,
                           feature_data: Union[np.ndarray, pd.DataFrame, List[List]],
                           target_data: Union[np.ndarray, pd.DataFrame, List] = None,
                           metadata: List[Dict] = None,
                           data_name: str = 'validation_data',
                           ) -> Dict:
        """Upload a subset of the training/validation data to be used for
            - Summary statistics for the model
            - To detect model drift
            - To use as baseline data for model interpretability
            - To seed new data in the client.

        Training and validation data with targets must be uploaded with the targets separate.

        Args:
            feature_data (Union[np.ndarray, pd.DataFrame, List[List]]): The feature data to upload.
            target_data (Union[np.ndarray, pd.DataFrame, List[List]]): The target data to upload.
            metadata (List[Dict]): Metadata key/value pairs for the dataset.
            data_name (str): The name of the data set.

        Returns:
            Dict: The status of the upload.
        """

        feature_data_np = coerce_data_as_numpy_array(feature_data)
        self.validate_feature_data_shape(feature_data_np)

        signed_s3_upload_post = api.signed_s3_upload_post(self.model_version_id, data_name)
        logger.debug(f'Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}')

        with tempfile.TemporaryDirectory() as data_temp_directory:
            data_temp_file = _build_h5_data_object(feature_data_np, target_data, metadata, data_temp_directory)
            files = {'file': (f'{data_name}.h5', open(data_temp_file, 'rb'))}
            form_fields = signed_s3_upload_post['form_fields']
            form_fields['AWSAccessKeyId'] = form_fields.pop('aws_access_key_id')
            s3_key = form_fields['key']
            requests.post(signed_s3_upload_post['url'], data=form_fields, files=files)
        return api.register_data_for_model(s3_key, self.model_version_id, data_name)

    def deploy_explainer(self,
                         explainer_name: str,
                         feature_data: Union[np.ndarray, pd.DataFrame, List[List]] = None,
                         target_data: Union[np.ndarray, pd.DataFrame, List] = None,
                         metadata: List[Dict] = None,
                         data_name: str = 'explainer_background_data.h5',
                         ) -> Dict:
        raises_invalid_explainer_configuration(
            explainer_name,
            self.model_config['model'].get('model_framework', ''),
            self.model_config['model'].get('model_type', ''),
            feature_data,
            self.model_config['model_features'].get('feature_names', '')
        )

        if feature_data is not None:
            feature_data_np = coerce_data_as_numpy_array(feature_data)
            self.validate_feature_data_shape(feature_data_np)
        else:
            feature_data_np = np.array([])

        signed_s3_upload_post = api.signed_s3_upload_post(self.model_version_id, data_name)
        logger.debug(f'Signed s3 upload post:\n{json.dumps(signed_s3_upload_post, indent=4)}')

        with tempfile.TemporaryDirectory() as data_temp_directory:
            data_temp_file = _build_h5_data_object(feature_data_np, target_data, metadata, data_temp_directory)
            files = {'file': (f'{data_name}.h5', open(data_temp_file, 'rb'))}
            form_fields = signed_s3_upload_post['form_fields']
            form_fields['AWSAccessKeyId'] = form_fields.pop('aws_access_key_id')
            s3_key = form_fields['key']
            requests.post(signed_s3_upload_post['url'], data=form_fields, files=files)
        return BasetenDeployedExplainer(
            api.deploy_explainer(s3_key, self.model_version_id, data_name, explainer_name)['id']
        )

    def validate_feature_data_shape(self, feature_data: np.ndarray):
        feature_data_shape = self.feature_shape(feature_data)
        model_input_shape = api.model_version_input_shape(self.model_version_id)
        if model_input_shape != feature_data_shape:
            raise SampleDataInputShapeError(f'The model has shape {model_input_shape}'
                                            f' but {feature_data_shape} provided for sample data.')

    def feature_shape(self, feature_data: np.ndarray) -> List:
        return list(feature_data.shape)[1:]

    @property
    def _model_config_path(self):
        return f'{BASETEN_CONFIG_DIRECTORY}/{self.model_version_id}/'

    @property
    def _model_config_file(self):
        return f'{self._model_config_path}model_config.yml'

    @property
    def model_config(self):
        if self._model_config is None:
            try:
                with open(self._model_config_file, 'r') as f:
                    self._model_config = yaml.safe_load(f)
            except Exception as e:
                logger.error(f'Unable to load model config for model_version_id {self.model_version_id} due to {e}')
                self._model_config = {}
        return self._model_config

    def generate_model_template(self,
                                model_type: str = None,
                                model_framework: str = None,
                                feature_names: List[str] = None,
                                class_labels: List[str] = None,
                                input_type: str = 'structured_data'):
        # TODO: validations
        data = {
            'model_type': model_type,
            'model_framework': model_framework,
            'input_type': input_type,
            'feature_names': feature_names or [],
            'class_labels': class_labels or []
        }

        if not os.path.exists(self._model_config_path):
            os.makedirs(self._model_config_path)
        with open(self._model_config_file, 'w') as f:
            f.writelines(model_template.render(data))
        return self._model_config_path



import functools
import json
import logging
from typing import Any, Dict, List

import keyring
import requests

from baseten.common import settings
from baseten.common.core import (KEYRING_SERVICE_NAME, ApiError,
                                 AuthorizationError)
from baseten.common.util import base64_encoded_json_str

logger = logging.getLogger(__name__)


def with_api_key(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        api_key = keyring.get_password(KEYRING_SERVICE_NAME, username=None)
        if not api_key:
            raise AuthorizationError('You must first run the `baseten login` cli command.')
        result = func(api_key, *args, **kwargs)
        return result
    return wrapper


@with_api_key
def models(api_key):
    query_string = '''
    {
      models {
        id,
        name
      }
    }
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']


@with_api_key
def create_model(api_key, model_name):
    if not model_name:
        import coolname
        model_name = coolname.generate_slug(2)
    query_string = f'''
    mutation {{
      create_model(name: "{model_name}") {{
        id,
        name
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model']


@with_api_key
def models_versions(api_key, model_id):
    query_string = f'''
    {{
      model(id: "{model_id}") {{
        versions {{
          id,
          s3_key,
          feature_names,
          feature_summary
        }}
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model']


@with_api_key
def model_version_input_shape(api_key, model_version_id):
    query_string = f'''
    {{
      model_version(id: "{model_version_id}") {{
        input_shape
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['model_version']['input_shape']


@with_api_key
def create_model_version(api_key,
                         model_id,
                         s3_key,
                         model_framework,
                         model_type,
                         input_shape,
                         feature_names,
                         feature_summary,
                         class_labels,
                         model_framework_req):
    encoded_input_shape = base64_encoded_json_str(input_shape)
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_feature_summary = base64_encoded_json_str(feature_summary)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    encoded_model_framework_req = base64_encoded_json_str(model_framework_req)
    query_string = f'''
    mutation {{
      create_model_version(model_id: "{model_id}",
                           s3_key: "{s3_key}",
                           model_framework: "{model_framework}",
                           model_type: "{model_type}",
                           encoded_input_shape: "{encoded_input_shape}",
                           encoded_feature_names: "{encoded_feature_names}",
                           encoded_feature_summary: "{encoded_feature_summary}",
                           encoded_class_labels: "{encoded_class_labels}",
                           encoded_model_framework_req: "{encoded_model_framework_req}") {{
        id,
        s3_key,
        feature_names,
        feature_summary
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_model_version']


@with_api_key
def create_data_transform(api_key, model_version_id, transformation_data, view_config):
    encoded_transformation_data = base64_encoded_json_str(transformation_data)
    encoded_view_config = base64_encoded_json_str(view_config)
    query_string = f'''
    mutation {{
      create_data_transform(model_version_id: "{model_version_id}",
                            encoded_transformation_data: "{encoded_transformation_data}",
                            encoded_view_config: "{encoded_view_config}") {{
        id
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_data_transform']


@with_api_key
def update_or_create_workflow(api_key, workflow_name, worklet_configs, query_configs):
    encoded_query_configs = base64_encoded_json_str(query_configs)
    encoded_worklet_configs = base64_encoded_json_str(worklet_configs)
    query_string = f'''
    mutation {{
      update_or_create_workflow(workflow_name: "{workflow_name}",
                                encoded_worklet_configs: "{encoded_worklet_configs}",
                                encoded_query_configs: "{encoded_query_configs}") {{
        id
        name
        worklets {{
            id
            name
        }}
        queries {{
            id
            name
        }}
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_or_create_workflow']


@with_api_key
def workflow(api_key, name):
    query_string = f'''
    {{
      workflow(name: "{name}") {{
        id
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['workflow']


@with_api_key
def update_workflow(api_key, workflow_id, workflow_name, worklet_configs, query_configs):
    encoded_worklet_configs = base64_encoded_json_str(worklet_configs)
    encoded_query_configs = base64_encoded_json_str(query_configs)
    query_string = f'''
    mutation {{
      update_workflow(workflow_id: "{workflow_id}",
                      name: "{workflow_name}",
                      encoded_worklet_configs: "{encoded_worklet_configs}",
                      encoded_query_configs: "{encoded_query_configs}") {{
        id
        name
        worklets {{
            id
            name
        }}
        queries {{
            id
            name
        }}
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_workflow']


@with_api_key
def create_workflow_action(api_key, workflow_id, action_name, workflow_action_config):
    encoded_workflow_action_data = base64_encoded_json_str(workflow_action_config)
    query_string = f'''
    mutation {{
      create_workflow_action(workflow_id: "{workflow_id}",
                      name: "{action_name}",
                      encoded_workflow_action_data: "{encoded_workflow_action_data}") {{
        id
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_workflow_action']


@with_api_key
def publish_workflow(api_key, workflow_version_id: str) -> Dict:
    query_string = f'''
    mutation {{
      publish_workflow_version(id: "{workflow_version_id}") {{
        id
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['publish_workflow_version']


@with_api_key
def transform_data(api_key, data_transform_id: str, input_data: Dict) -> Dict:
    predict_url = f'{settings.API_URL_BASE}/data_transforms/{data_transform_id}/invoke'
    resp = _post_rest_query(api_key, predict_url, {'input_data': input_data})
    resp.raise_for_status()
    return json.loads(resp.content)


@with_api_key
def update_transform_view(api_key, data_transform_id: str, view_config: Dict) -> Dict:
    encoded_view_config = base64_encoded_json_str(view_config)
    query_string = f'''
    mutation {{
      update_feature_transform_view(data_transform_id: "{data_transform_id}",
                                    encoded_view_config: "{encoded_view_config}") {{
        view_config
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_feature_transform_view']


@with_api_key
def signed_s3_upload_post(api_key, model_id, model_file_name):
    query_string = f'''
    {{
      signed_s3_upload_url(model_id: "{model_id}", model_file_name: "{model_file_name}") {{
        url,
        form_fields {{
          key,
          aws_access_key_id,
          policy,
          signature,
        }}
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['signed_s3_upload_url']


@with_api_key
def register_data_for_model(api_key, s3_key, model_version_id, data_name):
    query_string = f'''
    mutation {{
        create_sample_data_file(model_version_id: "{model_version_id}",
                                name: "{data_name}",
                                s3_key: "{s3_key}") {{
          id
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_sample_data_file']


@with_api_key
def deploy_explainer(api_key, s3_key, model_version_id, data_name, explainer_name):
    query_string = f'''
    mutation {{
        deploy_explainer(model_version_id: "{model_version_id}",
                                name: "{data_name}",
                                s3_key: "{s3_key}",
                                explainer_name: "{explainer_name}"),{{
          id
        }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['deploy_explainer']


@with_api_key
def predict_for_model(api_key, model_id: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    """Call the model's predict given the input json.

    Args:
        api_key (str)
        model_id (str)
        inputs (list)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    predict_url = f'{settings.API_URL_BASE}/models/{model_id}/predict'
    return _predict(api_key, predict_url, inputs, metadata)


@with_api_key
def predict_for_model_version(api_key, model_version_id: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    """Call the model version's predict given the input json.

    Args:
        api_key (str)
        model_version_id (str)
        inputs (list)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    predict_url = f'{settings.API_URL_BASE}/model_versions/{model_version_id}/predict'
    return _predict(api_key, predict_url, inputs, metadata)


@with_api_key
def explain_for_version(api_key, explainer_version_id: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    """Call the explainer version given the input json.

    Args:
        api_key (str)
        explainer_version_id (str)
        inputs (list)
        metadata (List[Dict]): Metadata key/value pairs (e.g. name, url), one for each input.

    Raises:
        RequestException: If there was an error communicating with the server.
    """
    explain_url = f'{settings.API_URL_BASE}/explainer_versions/{explainer_version_id}/explain'
    resp = _post_rest_query(api_key, explain_url, {'inputs': inputs, 'metadata': metadata})
    resp_json = json.loads(resp.content)
    return resp_json


@with_api_key
def set_primary(api_key, model_version_id: str):
    """Promote this version of the model as the primary version.

    Args:
        api_key (str)
        model_version_id (str)
    """
    query_string = f'''
    mutation {{
      update_model_version(model_version_id: "{model_version_id}", is_primary: true) {{
        id,
        is_primary,
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_model_version']


@with_api_key
def update_model_config(api_key, model_version_id: str, feature_names: list, class_labels: list = None):
    """Update the feature names for the model.

    Args:
        api_key (str)
        model_version_id (str)
        feature_names (list)
        class_labels (Optional[list]): applies only to classifiers.
    """
    encoded_feature_names = base64_encoded_json_str(feature_names)
    encoded_class_labels = base64_encoded_json_str(class_labels)
    query_string = f'''
    mutation {{
      update_model_version(model_version_id: "{model_version_id}",
                           encoded_feature_names: "{encoded_feature_names}",
                           encoded_class_labels: "{encoded_class_labels}") {{
        id,
        feature_names,
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['update_model_version']


@with_api_key
def invoke_worklet(api_key, workflow_id, worklet_name: str, inp, create_workflow_instance=False, dry_run=False):
    url = f'{settings.API_URL_BASE}/workflows/{workflow_id}/worklets/{worklet_name}/invoke'
    resp = _post_rest_query(api_key, url, {
        'worklet_input': inp,
        'dry_run': dry_run,
        'create_workflow_instance': create_workflow_instance,
    })
    resp_json = json.loads(resp.content)
    return resp_json


@with_api_key
def dryrun_atom(api_key, atom_name: str, atom_conf: dict, atom_input: Any, workflow_name: str):
    url = f'{settings.API_URL_BASE}/workflows/atom/run'
    resp = _post_rest_query(api_key, url, {
        'atom_name': atom_name,
        'atom_conf': atom_conf,
        'atom_input': atom_input,
        'workflow_name': workflow_name,
        'dry_run': True,
    })
    return json.loads(resp.content)


@with_api_key
def add_external_connection(api_key, connection_id, connection_uri):
    query_string = f'''
    mutation {{
      create_external_connection(connection_id: "{connection_id}",
                                 connection_uri: "{connection_uri}") {{
        connection_id
        connection_type
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_external_connection']


@with_api_key
def external_connections(api_key):
    query_string = f'''
    {{
      external_connections {{
        connection_id
        connection_type
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['external_connections']


@with_api_key
def install_requirements(api_key, requirements_txt):
    escaped_requirements_txt = requirements_txt.replace('\n', '\\n')  # Otherwise the mutation becomes invalid graphql.
    query_string = f'''
    mutation {{
      create_pynode_requirement(requirements_txt: "{escaped_requirements_txt}") {{
        id
        status
        error_message
      }}
    }}
    '''
    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['create_pynode_requirement']


@with_api_key
def requirement_status(api_key, requirement_id):
    query_string = f'''
    {{
      pynode_requirement(id: "{requirement_id}") {{
        id
        status
        error_message
      }}
    }}
    '''

    resp = _post_graphql_query(api_key, query_string)
    return resp['data']['pynode_requirement']


def _predict(api_key, predict_url: str, inputs: list, metadata: List[Dict] = None) -> List[List]:
    resp = _post_rest_query(api_key, predict_url, {'inputs': inputs, 'metadata': metadata})
    resp_json = json.loads(resp.content)
    return resp_json['predictions']


def _headers(api_key):
    return {'Authorization': f'Api-Key {api_key}'}


def _post_graphql_query(api_key, query_string) -> dict:
    resp = requests.post(f'{settings.API_URL_BASE}/graphql/', data={'query': query_string}, headers=_headers(api_key))
    if not resp.ok:
        logger.error(f'GraphQL endpoint failed with error: {resp.content}')
        resp.raise_for_status()
    resp_dict = resp.json()
    errors = resp_dict.get('errors')
    if errors:
        raise ApiError(errors[0]['message'], resp)
    return resp_dict


def _post_rest_query(api_key, url, post_body_dict):
    resp = requests.post(url, json=post_body_dict, headers=_headers(api_key))
    resp.raise_for_status()
    return resp



from __future__ import annotations

from typing import Any, List

from colorama import Fore

from baseten.common import api
from baseten.common.core import InvokingUnpublishedWorkflowError
from baseten.common.settings import API_URL_BASE
from baseten.common.util import print_error_response
from baseten.workflow.atom import RegisteredAtom
from baseten.workflow.query import Query
from baseten.workflow.worklet import Worklet


class Workflow:
    """Workflow is a meaningful flow of activities.

    Consists of worklets, their triggers and state.
    """

    def __init__(self, name: str):
        self._id = None
        self._name = name
        self._worklets = []
        self._queries = []

    def add_worklet(self, worklet: Worklet):
        self._worklets.append(worklet)

    def add_worklets(self, worklets: List[Worklet]):
        self._worklets.extend(worklets)

    def add_query(self, query: Query):
        self._queries.append(query)

    def add_queries(self, queries: List[Query]):
        self._queries.extend(queries)

    def publish(self):
        worklet_configs = [w.to_json() for w in self._worklets]
        query_configs = [q.to_json() for q in self._queries]
        response = api.update_or_create_workflow(self._name, worklet_configs, query_configs)
        self._id = response['id']
        print(Fore.GREEN + f'Published successfully. Visit: ' + Fore.WHITE + f'{API_URL_BASE}/workflows/{self._id}')

    def invoke_worklet(self,
                       worklet_name: str,
                       worklet_input,
                       create_workflow_instance=False,
                       dry_run=False,
                       print_response=True):
        if not self._id:
            raise InvokingUnpublishedWorkflowError('You must first publish the workflow before invoking its worklets.')

        response = api.invoke_worklet(self._id,
                                      worklet_name,
                                      worklet_input,
                                      create_workflow_instance,
                                      dry_run)

        if print_response:
            if not response['success']:
                print(Fore.RED + f'Invoking worklet "{worklet_name}" failed.')
                print_error_response(response)
            else:
                print(Fore.GREEN + f'Worklet {worklet_name} executed successfully with input:'
                      + Fore.WHITE + f' {worklet_input}')
                print(Fore.GREEN + f'Worklet Output:' + Fore.WHITE + f' {response["worklet_output"]}')
                if dry_run:
                    print(Fore.YELLOW + f'Execution Log:\n{response["execution_log"]}')

        return response

    def dryrun(self, reg_atom: RegisteredAtom, atom_input: Any):
        """
        Dryrun an atom.

        Dryrun happens on the backend where appropriate context is provided to the atom. Note
        that workflow_instance_id in the context will be none as that's not available at the
        atom level.
        Dryrun means that no state is implicitly persisted but Atom code may still create side
        effects. To guard against those, Atom code should use the dryrun flag which will be set
        to True for dry run but is False for regular execution.
        """
        name = reg_atom.name
        conf = reg_atom.conf
        response = api.dryrun_atom(name, conf, atom_input, self._name)
        if not response['success']:
            print(Fore.RED + f'Invoking atom "{name}" failed.')
            print_error_response(response)
        else:
            print(Fore.GREEN + f'Atom {name} executed successfully with input:'
                  + Fore.WHITE + f' {atom_input}')
            print(Fore.GREEN + f'Atom Output:' + Fore.WHITE + f' {response["atom_output"]}')
            print(Fore.YELLOW + f'Execution Log:\n{response["execution_log"]}')
        return response

    @classmethod
    def from_name(cls, workflow_name: str) -> Workflow:
        response = api.workflow(workflow_name)
        workflow = cls(workflow_name)
        workflow._id = response['id']
        return workflow



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['baseten', 'baseten.common', 'baseten.examples', 'baseten.workflow']

package_data = \
{'': ['*'], 'baseten': ['templates/*']}

install_requires = \
['boto3>=1.9,<2.0',
 'click>=7.0,<8.0',
 'cloudpickle==1.6.0',
 'colorama>=0.4.3,<0.5.0',
 'coolname>=1.1.0,<2.0.0',
 'h5py>=2.10.0,<3.0.0',
 'jinja2>=2.10.3,<3.0.0',
 'joblib>=0.13.2,<0.14.0',
 'keyring>=19.2,<20.0',
 'libcst>=0.3.4,<0.4.0',
 'pandas>=0.25.1,<0.26.0',
 'requests>=2.22,<3.0',
 'tensorflow>=2.3.0,<3.0.0']

entry_points = \
{'console_scripts': ['baseten = baseten.cli:cli_group']}

setup_kwargs = {
    'name': 'baseten',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Amir Haghighat',
    'author_email': 'amir@baseten.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
